#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""
## Metadata Store

The configuration of each pipeline, step, backend, and produced artifacts are
all tracked within the metadata store. The metadata store is an SQL database,
and can be `sqlite` or `mysql`.

Metadata are the pieces of information tracked about the pipelines, experiments
and configurations that you are running with ZenML. Metadata are stored
inside the metadata store.
"""

from zenml.metadata_stores.base_metadata_store import BaseMetadataStore
from zenml.metadata_stores.mysql_metadata_store import MySQLMetadataStore
from zenml.metadata_stores.sqlite_metadata_store import SQLiteMetadataStore

__all__ = [
    "BaseMetadataStore",
    "MySQLMetadataStore",
    "SQLiteMetadataStore",
]
